@@ -18,6 +18,7 @@ from scipy.stats import multivariate_normal as mv_norm
 from matplotlib import pyplot as plt, axes
 from mpctools.extensions import npext
 import seaborn as sns
+import pandas as pd
 import numpy as np
 import warnings
 
@@ -28,12 +29,13 @@ def plot_matrix(
     min_max=None,
     show_val=False,
     ax=None,
-    cbar=None,
-    labels=None,
+    cax=None,
+    x_labels=None,
     y_labels=None,
     x_rot=0,
     y_rot=0,
     fmt=".2f",
+    fs=10,
 ):
     """
     Draw Hinton/Heatmap diagram for visualizing a weight matrix.
@@ -45,6 +47,8 @@ def plot_matrix(
     Rows (first-dimension) of the matrix are printed along the y-axis, with columns (2nd dimension) along the X-axis.
     Note that by default, the y-axes is inverted.
 
+    TODO: Support NaN plotting
+
     Inspired by [Hinton Demo](https://matplotlib.org/examples/specialty_plots/hinton_demo.html). The heatmap makes
     use of seaborn functionality, with some pre/post processing.
 
@@ -59,13 +63,14 @@ def plot_matrix(
                             Tuple/List: (minimum, maximum). Only for Heatmaps.
     :param show_val:    If True, then show the numerical value on the heatmap/hinton
     :param ax:          Axes to plot on
-    :param cbar:        Only relevant when mode is 'heatmap': if not None, specifies a seperate axes for the colour bar.
-                        If False, then do not plot a heatmap.
-    :param labels:      Labels for the axes.
-    :param y_labels:    If not None, then use separate labels for the y-axis.
+    :param cax: Only relevant when mode is 'heatmap': if not None, specifies a seperate  axes for the colour bar.
+                If False, then do not plot a heatmap colour-bar.
+    :param x_labels:    Labels for the x-axes.
+    :param y_labels:    Labels for the y-axes - if None, use same as x
     :param x_rot:       Rotation for the X-Axis Labels
     :param y_rot:       Rotation for the Y-Axis Labels
     :param fmt:         Formatting String for Value labels (if any)
+    :param fs: Font-Size
     :return:
     """
     # Sort out the mode
@@ -127,6 +132,7 @@ def plot_matrix(
                     horizontalalignment="center",
                     verticalalignment="center",
                     color="black" if w > 0 else "white",
+                    fontsize=fs,
                 )
         ax.set_ylim(-1, matrix.shape[0])
         ax.set_xlim(-1, matrix.shape[1])
@@ -139,26 +145,27 @@ def plot_matrix(
             annot=show_val,
             fmt=fmt,
             ax=ax,
-            cbar=cbar is not False,
-            cbar_ax=cbar if isinstance(cbar, axes.Axes) else None,
+            cbar=cax is not False,
+            cbar_ax=cax if isinstance(cax, axes.Axes) else None,
         )
 
     # Add Ticks/Labels
-    if labels is not None:
-        y_labels = labels if y_labels is None else y_labels
+    y_labels = x_labels if (type(y_labels) is str and y_labels.lower() == 'same') else y_labels
+    if x_labels is not None:
         if mode:
-            ax.set_xticks(np.arange(len(labels)))
+            ax.set_xticks(np.arange(len(x_labels)))
         else:
-            ax.set_xticks(np.arange(0.5, len(labels) + 0.5))
+            ax.set_xticks(np.arange(0.5, len(x_labels) + 0.5))
         ax.set_xticklabels(
-            labels, rotation=x_rot, horizontalalignment="center" if x_rot == 0 else "right",
+            x_labels, rotation=x_rot, horizontalalignment="center" if x_rot == 0 else "right", fontsize=fs
         )
+    if y_labels is not None:
         if mode:
             ax.set_yticks(np.arange(len(y_labels)))
         else:
             ax.set_yticks(np.arange(0.5, len(y_labels) + 0.5))
         ax.set_yticklabels(
-            y_labels, rotation=y_rot, verticalalignment="center" if y_rot == 0 else "bottom",
+            y_labels, rotation=y_rot, verticalalignment="center" if y_rot == 0 else "bottom", fontsize=fs
         )
 
 
@@ -293,7 +300,76 @@ def plot_categorical(
     return plot, colorbar
 
 
-def plot_contour(axs=None, x=None, y=None, kind="G", params=(np.zeros(2), np.eye(2)), res=100):
+def plot_lifespans(
+    ts: np.ndarray, colors=False, ax=None, y_labels=None, x_labels=None, fs=15, invert_y=True,
+):
+    """
+    Plot life-spans of time-series in the format of a gantt-chart.
+
+    :param ts: A 2D array to plot, where time is along the rows (to follow from Pandas
+                        DataFrames). The values themselves are generally ignored (although see
+                        colors), and visibility is governed by it being a finite number.
+    :param colors:      Colours to use for the bars. Can be:
+                          False: Default - Do not colour (all gray)
+                          True: Interpret the values in time_series as colours
+                          List/Tuple: List of colours to iterate over different series
+                          str: Named colour map
+    :param ax:          If provided, uses this axes
+    :param y_labels:    Labels to use for the y-value
+    :param x_labels:    Labels to use for the x-axis
+    :param fs:          Font-Size
+    :param invert_y: If True (default) invert y-axis
+    :return:
+    """
+    # Evaluate Parameters
+    n_time, n_series = ts.shape
+    ax = plt.gca() if ax is None else ax
+
+    # Generate Discrete Colour Map
+    if type(colors) in (list, tuple):
+        color_list = colors
+    elif type(colors) is str:
+        clrmap = plt.cm.get_cmap(colors)
+        color_list = clrmap(np.linspace(0, 1, min(n_series, clrmap.N)))
+    elif colors is False:
+        color_list = [(0.75, 0.75, 0.75, 1.0)] # Gray Colour
+    else:
+        color_list = None
+
+    # To simplify computation load
+    if color_list is not None:
+        ts = pd.notnull(ts).astype(float)
+        ts[ts == 0] = np.NaN
+
+    # Iterate over time-series (columns):
+    for i in range(n_series):
+        l, v, p = npext.run_lengths(ts[:, i], how='I', return_values=True, return_positions=True)
+        runs = [z for z in zip(p, l)]
+        cols = v if color_list is None else color_list[i % len(color_list)]
+        ax.broken_barh(runs, (i + 0.1, 0.8), facecolors=cols)
+
+    # Set Labelling
+    ax.set_yticks(np.arange(0.5, n_series, 1.0))
+    if y_labels is not None:
+        ax.set_yticklabels(y_labels)
+    if x_labels is not None:
+        x_ticks = ax.get_xticks().astype(int)
+        x_ticks = x_ticks[np.logical_and(x_ticks < n_time, x_ticks > -1)]
+        ax.set_xticks(x_ticks)
+        ax.set_xticklabels(np.asarray(x_labels)[x_ticks])
+    ax.tick_params(labelsize=fs)
+    ax.grid(True, axis="x")
+    if invert_y:
+        ax.invert_yaxis()
+
+    # Return the Axes
+    return ax
+
+
+def plot_contour(
+        axs=None, x=None, y=None, kind="G", params=(np.zeros(2), np.eye(2)), res=100,
+        show_means=200, levels=None
+):
     """
     Generate a contour plot
 
@@ -307,6 +383,8 @@ def plot_contour(axs=None, x=None, y=None, kind="G", params=(np.zeros(2), np.eye
                   the value at each pt
     :param params: Any additional parameters to use when using the Gaussian plot
     :param res:   The resolution along the X and Y.
+    :param show_means: If not None, display the Mean as a cross of specified size
+    :param levels: Number of levels to show (in axes.contour())
     :return:      The plot parameters as returned by pyplot.contour()
     """
     # Handle specified options
@@ -324,6 +402,11 @@ def plot_contour(axs=None, x=None, y=None, kind="G", params=(np.zeros(2), np.eye
 
     # Handle Kind
     if type(kind) == str and kind.lower() == "g":
-        axs.contour(X, Y, mv_norm.pdf(np.stack((X, Y), axis=-1), *params))
+        if levels is not None:
+            axs.contour(X, Y, mv_norm.pdf(np.stack((X, Y), axis=-1), *params), levels)
+        else:
+            axs.contour(X, Y, mv_norm.pdf(np.stack((X, Y), axis=-1), *params))
+        if show_means is not None:
+            axs.scatter([params[0][0]], [params[0][1]], s=show_means, marker='x')
     else:
         raise NotImplementedError("Executable Kind is not yet implemented")
