@@ -13,10 +13,14 @@ see http://www.gnu.org/licenses/.
 Author: Michael P. J. Camilleri
 """
 
-from scipy.optimize import linear_sum_assignment as hungarian
+from scipy.spatial.distance import pdist
+from hotelling import stats as hstats
+from lapsolver import solve_dense
+from deprecated import deprecated
+from scipy.stats import entropy, f
 from scipy.special import gamma
-from scipy.stats import entropy
 from functools import reduce
+import pandas as pd
 import numpy as np
 import itertools
 
@@ -114,7 +118,7 @@ def exp10(a):
 def clamp(a, cutoff, limits):
     """
     Clamp the values in a to limits at the cutoff: specifically, values less than or equal to the
-    cutoff are converted to limits[0] and values greather than cutoff are converted to limits[1].
+    cutoff are converted to limits[0] and values greater than cutoff are converted to limits[1].
     This is a counterpart to the clip function.
 
     :param a:       Array
@@ -145,7 +149,7 @@ def mad(a, b, axis=None, keepdims=False):
 
 def multiply(arrays):
     """
-    Multiply (element-wise) a series of arrays: they must all abe of the same shape!
+    Multiply (element-wise) a series of arrays: they must all be of the same shape!
 
     :param arrays:  Tuple/list of arrays to multiply
     :return:        Element-wise multiple
@@ -203,26 +207,26 @@ def run_lengths(a, how="I", return_values=False, return_positions=False):
     """
     # Prepare
     #  We fist compute assuming that we are going to use everything! This allows us to have correct
-    #  position information We also need to convert arrays of NaN to lists so that NaN works.
+    #  position information. We also need to convert arrays of NaN to lists so that NaN works.
     #  This will slow down but cannot be helped.
     if type(a) == np.ndarray:
-        a = [i if not np.isnan(i) else np.nan for i in a.flatten()]
+        a = [i if pd.notnull(i) else np.NaN for i in a.flatten()]
     rls = np.asarray(
-        [(sum(1 for _ in l), n) for n, l in itertools.groupby(a)]
+        [(sum(1 for _ in l), n) for n, l in itertools.groupby(a)], dtype="object",
     )  # Now, compute run-lengths and types.
     pos = (
-        np.asarray([0, *np.cumsum(rls[:, 0])[:-1]], dtype=int) if len(rls) > 0 else []
+        np.asarray([0, *np.cumsum(rls[:, 0].astype(int))[:-1]], dtype=int) if len(rls) > 0 else []
     )  # Compute positions just in case.
     # Compute: Branch on Format
     if how.lower() == "a":
         lens, vals = (rls[:, 0], rls[:, 1]) if len(rls) > 0 else ([], [])
     elif how.lower() == "i":
-        n_nan = ~np.isnan(rls[:, 1])
+        n_nan = pd.notnull(rls[:, 1])
         lens, vals, pos = (
             (rls[n_nan, 0], rls[n_nan, 1], pos[n_nan]) if len(n_nan) > 0 else ([], [], [])
         )
     elif how.lower() == "o":
-        nan = np.isnan(rls[:, 1])
+        nan = pd.isna(rls[:, 1])
         lens, vals, pos = (rls[nan, 0], rls[nan, 1], pos[nan]) if len(nan) > 0 else ([], [], [])
     else:
         raise ValueError('Incorrect format for "how" parameter.')
@@ -235,12 +239,16 @@ def run_lengths(a, how="I", return_values=False, return_positions=False):
     return tuple(to_return) if len(to_return) > 1 else to_return[0]
 
 
-def array_nan_equal(left, right):
+def array_nan_equal(left, right, axis=None):
     """
     Compares two numpy arrays to test for equality, but considers np.nan values to be equal to each
     other
+
     :param left: The Left array to consider
     :param right: The Right array to consider
+    :param axis: Optional: Dimension along which to compare - note this only affects the equality in
+                 terms of values: i.e. if arrays are not the same shape, then only a single value is
+                 returned nonetheless.
     :return: True if left is the same shape as right, and all elements are equal (including nan's at
              same positions)
     """
@@ -255,10 +263,10 @@ def array_nan_equal(left, right):
         return False
 
     # Return
-    return bool(
-        np.logical_or(
-            np.asarray(left == right), np.logical_and(np.isnan(left), np.isnan(right))
-        ).all()
+    return (
+        np.logical_or(np.asarray(left == right), np.logical_and(pd.isnull(left), pd.isnull(right)))
+        .all(axis=axis)
+        .astype(bool)
     )
 
 
@@ -309,17 +317,59 @@ def make_diagonal(on_diag, off_diag, size):
     return np.eye(size) * on_diag + non_diag(np.full(shape=[size, size], fill_value=off_diag))
 
 
-def maximise_trace(x):
+def hungarian(costs: np.ndarray, maximise=False, cutoff=None, row_labels=None, col_labels=None):
     """
-    Maximise the Trace of a SQUARE Matrix X using the Hungarian Algorithm
+    Solves the Assignment problem.
 
-    :param x: Numpy 2D SQUARE Array
-    :return: Tuple containing (in order):
-                * optimal permutation of columns to achieve a maximal trace
-                * size of this trace
+    This method is a wrapper lapsolver's solve_dense that:
+     1. Can Threshold certain costs,
+     2. Can Handle np.NaN (as np.Inf)
+     3. Can deal with rows/columns of NaN
+     4. Can keep track of labels, rather than just indices
+
+    :param costs:      Cost Matrix to optimise
+    :param maximise:   (default: False) - Calculates a maximum weight matching if true.
+    :param cutoff:     If set, use this as a threshold. The cutoff range depends on whether
+                       maximising (in which case lower-values are invalidated) or minimising
+                       (higher values inadmissable).
+    :param row_labels: Row-Labels (optional) - If None, using 0-based indices
+    :param col_labels: Column-Labels (optional) - If None, using 0-based indices
+    :return:
     """
-    _rows, _cols = hungarian(np.full(len(x), np.max(x)) - x)
-    return _cols, x[_rows, _cols].sum()
+    # Prepare
+    _cost = costs.astype(float)
+
+    # Handle Edge Cases
+    _cost[~np.isfinite(_cost)] = np.NaN
+    if cutoff is not None:
+        if maximise:
+            _cost[_cost < cutoff] = np.NaN
+        else:
+            _cost[_cost > cutoff] = np.NaN
+
+    # Extract only valid rows/columns (i.e. where the is at least one element which is valid)
+    valid = np.isfinite(_cost)
+    if ~valid.any():  # Guard against having no valid assignments
+        return [], []
+    val_r, val_c = valid.any(axis=1), valid.any(axis=0)
+    _cost = _cost[val_r, :]
+    _cost = _cost[:, val_c]
+
+    # Perform Hungarian (but handle Maximisation)
+    if maximise:
+        finite = np.isfinite(_cost)
+        _cost[finite] = np.max(_cost[finite]) - _cost[finite]
+    r, c = solve_dense(_cost)
+
+    # Map to original Indices
+    r_ids, c_ids = np.where(val_r)[0][r], np.where(val_c)[0][c]
+    if row_labels is not None:
+        r_ids = np.asarray(row_labels)[r_ids]
+    if col_labels is not None:
+        c_ids = np.asarray(col_labels)[c_ids]
+
+    # Return
+    return r_ids, c_ids
 
 
 def swap_columns(x, cols):
@@ -336,6 +386,7 @@ def swap_columns(x, cols):
     return temp
 
 
+@deprecated('Functionality can in general be done by way of np.array()')
 def ensure2d(a, axis=0):
     """
     Returns a matrix of dimensionality 2 from a potentially linear vector
@@ -344,7 +395,9 @@ def ensure2d(a, axis=0):
     :param axis: Axis to append if missing: either 0 or 1
     :return: 2D Matrix
     """
-    if np.ndim(a) > 1:
+    if np.ndim(a) > 2:
+        raise ValueError('Dimension must be 2 or less.')
+    elif np.ndim(a) == 2:
         return a
     elif np.ndim(a) == 1:
         return a[np.newaxis, :] if axis == 0 else a[:, np.newaxis]
@@ -362,6 +415,7 @@ class Dirichlet:
     An alternative Dirichlet Class which
         a) Precomputes the normalisation parameters for efficiency
         b) can vectorise over matrices, where the last-dimension is the probability distribution.
+        c) Handles 0's in the x's
     """
 
     def __init__(self, alpha, ignore_zeros=False):
@@ -375,15 +429,19 @@ class Dirichlet:
                              which are 1. This is based on the assumption that such dimensions
                              will always be zero and will be known in advance.
         """
-        alpha = np.asarray(alpha)  # K   Dimensional
+        alpha = np.asarray(alpha)  # K-Dimensional
         # Do some Checks
         if np.any(alpha <= 0):
             raise ValueError("All dimensions of Alpha must be greater than 0.")
-        self._alpha = alpha - 1.0  # K Dimensional
-        self.__zeros = alpha == 1
+        self._alpha_m1 = alpha - 1.0  # K Dimensional
+        self.__zeros = alpha == 1  # Find location of zeros
         if ignore_zeros:
-            self._alpha = np.ma.array(self._alpha, mask=self.__zeros)
+            # Check that not all masked:
+            if self.__zeros.all():
+                raise ValueError("If ignoring zeros, you cannot have all of Alpha = 1!")
+            # We need to mask both alpha and alpha-1:
             alpha = np.ma.array(alpha, mask=self.__zeros)
+            self._alpha_m1 = np.ma.array(self._alpha_m1, mask=self.__zeros)
         self.norm = np.prod(gamma(alpha), axis=-1) / gamma(
             np.sum(alpha, axis=-1)
         )  # K-1 Dimensional
@@ -406,7 +464,12 @@ class Dirichlet:
         #      we are taking product this does not matter
         #   b) If we are ignoring zeros, this is masked by definition, which means that the result
         #      of the power is masked and so is the product!
-        return np.prod(np.power(x, self._alpha), axis=-1) / self.norm
+        # However, just in case in the future we decide to support all ones when ignoring,
+        # then we check for all zeros:
+        if self.__zeros.all():
+            return 1 / self.norm
+        else:
+            return np.prod(np.power(x, self._alpha_m1), axis=-1) / self.norm
 
     def logpdf(self, x):
         """
@@ -419,15 +482,17 @@ class Dirichlet:
         # Ensure that it is non-zero where zeros are not allowed
         if np.any(x[~self.__zeros] <= 0):
             raise ValueError("0-valued probabilities are only allowed where alpha=1")
-        # In this case, we must mask to avoid the issues with log since -inf * 0 = NaN... however,
-        #   we also set the error state appropriately since the mask does not appear to work on
-        #   log...
-        # Note that while we may wish to support alpha=1 even when non-zero, in any case,
-        #   this will not have any effect since for alpha=1, _alpha=0 and hence,
-        #   the multiplication below will be with 0!
-        with np.errstate(divide="ignore"):
+        # In this case, we must mask to avoid the issues with log since -inf * mask = NaN and not 0
+        #   as we would expect. Since we are enforcing that 0-valued probabilities are only
+        #   allowed when alpha=1, then we can mask out all such values, because in any case,
+        #   _alpha_m1=0 (for all self.__zeros) and hence, the multiplication will be 0 in any case.
+        # We do have to handle the case where all of _alpha_m1 is 0, in which case,
+        #    the sum should be just 0
+        if self.__zeros.all():
+            return -self.lognorm
+        else:
             x = np.ma.array(x, mask=self.__zeros)
-            return np.sum(np.multiply(np.log(x), self._alpha), axis=-1) - self.lognorm
+            return np.sum(np.multiply(np.log(x), self._alpha_m1), axis=-1) - self.lognorm
 
     def logsumpdf(self, x):
         """
@@ -445,7 +510,7 @@ class Dirichlet:
 
         :return:    A single sample
         """
-        return self._recursive_sample(self._alpha + 1)
+        return self._recursive_sample(self._alpha_m1 + 1)
 
     @staticmethod
     def _recursive_sample(alpha):
@@ -460,6 +525,90 @@ class Dirichlet:
             return sample
 
 
+class delta:
+    """
+    By Abuse of what a PDF is, this is a spike distribution which has probability 1 at a specific
+    value, and 0 otherwise.
+    """
+
+    def __init__(self, loc):
+        self.loc = np.array(loc, ndmin=1)
+
+    def pdf(self, x):
+        return int(np.array_equal(self.loc, np.array(x, ndmin=1)))
+
+    def logpdf(self, x):
+        return 0 if self.pdf(x) == 1 else np.NINF
+
+
+def ttest_mult(a, b, equal_var=False):
+    """
+    Calculates the Multivariate Hotelling T^2 Statistic for two independent multivariate samples
+
+    This is a two-sided test for the null hypothesis that 2 independent samples have identical
+    average (expected) values. If not pooling covariance, the multivariate computation of the
+    Welch T2 statistic follows the distance formulation in [1], while the p-value computation then
+    uses the method described in [2] (based on the F-Distribution). Otherwise, (i.e. pooled
+    covariance) this acts as a wrapper around the hotelling package, to confirm to the scipy ttest
+    nomenclature.
+
+    Parameters
+    -----------
+    a, b : array_like
+      These must be 2-Dimensional arrays, with sample index along the rows (first dimension).
+    equal_var : bool
+      If true, assume a pooled Covariance Matrix: otherwise, follow the method in [1]/[2]
+
+    Returns
+    --------
+    statistic: float
+       The calculated T2 statistic
+    pvalue: float
+       The two-tailed p-value
+    dof: int
+        Degrees of Freedom
+
+    References
+    ----------
+    .. [1] Alexander V. Alekseyenko, Multivariate Welch t-test on distances, Bioinformatics,
+           Volume 32, Issue 23, 1 December 2016, Pages 3552–3558,
+           https://doi.org/10.1093/bioinformatics/btw524
+    .. [2] https://stackoverflow.com/questions/25412954/hotellings-t2-scores-in-python/59152294#59152294
+    """
+
+    # Compute some Sizes
+    if np.ndim(a) != 2 or np.ndim(b) != 2:
+        raise ValueError("Parameters a/b must be 2D!")
+    if a.shape[1] != b.shape[-1]:
+        raise ValueError("Data (a/b) must have the same number of features")
+    na, p = a.shape
+    nb = b.shape[0]
+
+    # Calculate DOF
+    dof = na + nb - p - 1
+
+    if equal_var:
+        # Just Wrap around Hotelling's package
+        t2, _, pvalue, _ = hstats.hotelling_t2(a, b, True)
+    else:
+        # Compute Distances as per [1] (not the most efficient, but correct)
+        da = np.square(pdist(a, "euclidean")).sum()
+        db = np.square(pdist(b, "euclidean")).sum()
+        dz = np.square(pdist(np.vstack([a, b]), "euclidean")).sum()
+        # Compute T2 Statistic as per [1] and [2]
+        t2 = (
+            ((na + nb) / (na * nb))
+            * (dz / (na + nb) - da / na - db / nb)
+            / (da / (na ** 2 * (na - 1)) + db / (nb ** 2 * (nb - 1)))
+        )
+        # Now compute p-Value as per [2]
+        t2 = t2 * dof / (p * (na + nb - 2))
+        pvalue = 1 - f.cdf(t2, p, dof)
+
+    # return
+    return t2, pvalue, dof
+
+
 def sum_to_one(x, axis=None, norm=False):
     """
     Ensure that the elements of x sum to 1 (normally for probabilities), by dividing by their sum.
@@ -560,25 +709,32 @@ def conditional_entropy_2D(emission, prior=None, base=None):
         )
 
 
-def mutual_information(prior, emission, base=None):
+def mutual_information(prior, emission, normalised=False, base=None):
     """
-    Compute the Mutual information between an input (Z) and set of output (X) variables, under the
+    Computes the Mutual information between an input (Z) and set of output (X) variables, under the
       assumption that when there is more than 1 X variable, they are conditionally independent of
-      each other given Z (i.e. the Naive Bayes assumption). Be careful however, that as the
-      number of X variables increases, the dimensionality of the problem explodes (since we need
-      to work with the full joint marginal)!
+      each other given Z (i.e. the Naive Bayes assumption). This allows the joint over the output
+      variables to be simply the outer product of their individual probabilities. Be careful
+      however, that as the number of X variables increases, the dimensionality of the problem
+      explodes (since we need to work with the full joint marginal)!
 
     Note (1) - This is NOT the Conditional Mutual Information, which is something different: this is
-      just the MI between a Z and a multi-dimensional X (taken as one variable).
+      just the MI between a Z and a (possibly multi-dimensional) output X (taken as one global
+      variable).
 
     Note (2) - This only supports a 1D latent variable (Z): i.e. there may be multiple (X) variables
       but only 1 (Z).
 
+    Note (3) - Normalisation happens according to the Entropy of the output X: this is because,
+      usually, we want to find out how well Y can explain X, and hence, it should reach 1 only
+      when it can fully explain it.
+
     :param prior:       Prior Distribution over Z [1D array]
     :param emission:    Conditional Distribution over X given Z. This can be either:
                             a) 2D Numpy array, with Z along the rows, for one conditional
                             b) List of 2D Numpy arrays, each constituting a 2D Numpy array (Z along
                                rows) showing the emission of a variable.
+    :param normalised: If True, normalise by the Entropy of X
     :param base:        The numeric base to operate with.
     :return:            Mutual Information
     """
@@ -597,11 +753,12 @@ def mutual_information(prior, emission, base=None):
                 ).ravel()  # We are continuously increasing in size...
         emission = np.asarray(pXZ)  # Will be array of size |Z| by |X_1|*|X_2|*...|X_n|
 
-    # Compute Marginal X:
-    pX = np.matmul(prior, emission)
+    # Compute Marginal X and its entropy
+    hX = entropy(np.matmul(prior, emission), base=base)
 
     # Now Compute Entropies and return
-    return entropy(pX, base=base) - conditional_entropy(emission=emission, prior=prior, base=base)
+    mi = hX - conditional_entropy(emission=emission, prior=prior, base=base)
+    return mi / hX if normalised else mi
 
 
 def markov_stationary(transition):
