@@ -13,6 +13,8 @@ see http://www.gnu.org/licenses/.
 Author: Michael P. J. Camilleri
 """
 
+from datetime import timedelta, datetime
+from itertools import tee, islice
 import numpy as np
 import shutil
 import copy
@@ -24,6 +26,101 @@ import os
 ################################################
 
 
+class BitField:
+    """
+    A Class to define a bit-field
+    """
+    def __init__(self, width, vals=()):
+        """
+        Initialise with the list of values set to True and a specified width.
+        :param width: The size of the flag
+        :param vals: The set of 'on' values
+        """
+        if len(vals) > 0:
+            self.__flag = f"{(10 ** (width - 1 - np.asarray(vals, dtype=int))).sum():0{width}d}"
+        else:
+            self.__flag = "0" * width
+
+    def __repr__(self):
+        return self.__flag
+
+    def ison(self, item):
+        return self.__flag[item] == "1"
+
+    def __contains__(self, item):
+        return self.ison(item)
+
+    def __len__(self):
+        return len(self.__flag)
+
+
+class Pool:
+    """
+    Defines a pool of object assignments. By this we mean that it keeps track of a property for a
+    set of indices (hashables), and when indices die, can assign to new ones.
+    """
+
+    def __init__(self, values):
+        """
+        Creates the Pool object
+
+        :param values: The allowable values that can be returned. Must be an indexable iterable.
+        """
+        self.__values = values
+        self.__val_map = {}
+        self.__unused_val = {*range(len(values))}
+
+    def update(self, indices):
+        """
+        Updates the state with the new index-set
+
+        :param indices: New set of indices. This can be anything that is hashable.
+        :return: self, for chaining.
+        """
+        for _t_id in set(self.__val_map.keys()).difference(set(indices)):
+            self.__unused_val.add(self.__val_map.pop(_t_id))
+        for _t_id in set(indices).difference(set(self.__val_map.keys())):
+            self.__val_map[_t_id] = self.__unused_val.pop()
+        return self
+
+    def __getitem__(self, item):
+        return self.__values[self.__val_map[item]]
+
+    @property
+    def capacity(self):
+        return len(self.__values)
+
+    @property
+    def values(self):
+        return self.__values
+
+    def assigned(self):
+        """
+        Returns the assignment status of each value (much like a reverse dict).
+        :return: A Dictionary showing for each value the assigned index or None As per Python
+        Standard, the returned order follows the set of values as passed to the Initialiser.
+        """
+        reverse_dict = {v: None for v in self.__values}
+        for index, val_idx in self.__val_map.items():
+            reverse_dict[self.__values[val_idx]] = index
+        return reverse_dict
+
+
+def window(iterable, size):
+    """
+    Return a sliding window iterator over elements.
+
+    From https://stackoverflow.com/questions/6822725/rolling-or-sliding-window-iterator
+
+    :param iterable: An iterable
+    :param size:     Size of the sliding window
+    :return:         Iterator
+    """
+    itrs = tee(iterable, size)
+    shiftedStarts = [islice(anItr, s, None) for s, anItr in enumerate(itrs)]
+    return zip(*shiftedStarts)
+
+
 def to_tuple(value):
     """
     Static (Module) Method for converting a scalar into a tuple if not already (tuple/list/numpy
@@ -125,7 +222,7 @@ def pad(a, length, value):
     """
     Pad a list (not tuple). Note that if length is less than the len(a) then this has no effect.
 
-    :param a:       The original List
+    :param a:       The original List. Must be a list and NOT a numpy array!
     :param length:  The desired length
     :param value:   The Value to fill with
     :return:        Padded List. A copy is always returned!
@@ -157,6 +254,17 @@ def masked_list(l, m, mask_in=True):
     return lcpy
 
 
+def default(var, val):
+    """
+    Replaces var with val if var is None:
+
+    :param var: The variable to handle
+    :param val: The value to use if var is None
+    :return: var if var is not None, else val
+    """
+    return var if var is not None else val
+
+
 ################################################
 #            Printing & Formatting             #
 ################################################
@@ -213,6 +321,73 @@ def float_list(_list, prec=2):
     return ["{1:.{0}f}".format(prec, f) for f in _list]
 
 
+def time_list(_list, fmt="%H:%M:%S"):
+    """
+    Formats a list of time-points in 'HH:MM:SS'. Currently, only supports _list values in ms
+
+    :param _list:
+    :param fmt:  String format to employ
+    :return: String representation
+    """
+    s_l = []
+    for t in _list:
+        s_l.append((datetime(1970, 1, 1) + timedelta(milliseconds=t)).strftime(fmt))
+    return s_l
+
+
+def show_time(_time, minimal=True, ms=False):
+    """
+    Shows a time-point in the most appropriate manner (Days/Hours/Minutes/Seconds)
+
+    :param _time: Time (in seconds)
+    :param minimal: If True, and less than 60, show in minimal view (s.ms). Does not effect if
+                    more than 1 minute.
+    :param ms: If True, show millisecond precision
+    :return: String representation
+    """
+    if _time < 60 and minimal:
+        return f"{_time:.3f}s" if ms else f"{int(_time)}s"
+    else:
+        _time_str = datetime(1970, 1, 1) + timedelta(seconds=float(_time))
+        if _time < 3600:
+            return _time_str.strftime("%M:%S.%f")[:-3] if ms else _time_str.strftime("%M:%S")
+        elif _time < 86400:
+            return _time_str.strftime("%H:%M:%S.%f")[:-3] if ms else _time_str.strftime("%H:%M:%S")
+        else:
+            _time_str -= timedelta(hours=24)
+            return (
+                _time_str.strftime("%jD+%H:%M:%S.%f")[:-3]
+                if ms
+                else _time_str.strftime("%jD+%H:%M:%S")
+            )
+
+
+def int_list(_list, _sort=True):
+    """
+    Formats a list of unique integers, summarising ranges
+    :param _list: List to display
+    :param _sort: If True, sort the elements
+    :return:
+    """
+    _list = np.sort(_list).astype(int) if _sort else np.asarray(_list, dtype=int)
+    _diff = [*(np.diff(_list) > 1), True]
+
+    st = _list[0]
+    ranges = []
+    for el in np.argwhere(_diff):
+        nd = _list[el][0]
+        if st == nd:
+            ranges.append(f"{st}")
+        elif nd == st + 1:
+            ranges.append(f"{st}")
+            ranges.append(f"{nd}")
+        else:
+            ranges.append(f"{st}-{nd}")
+        st = _list[el + 1][0] if el < len(_list) - 1 else None
+
+    return ", ".join(ranges)
+
+
 def str_width(_iterable):
     """
     Returns the maximum width of all strings in iterable
@@ -252,7 +427,7 @@ def short_int(_int):
 
 
 ################################################
-#                OS Manipulation               #
+#            OS & Python Manipulation          #
 ################################################
 
 
@@ -279,3 +454,4 @@ def make_dir(_path, _clear=False):
     except OSError:
         if not os.path.isdir(_path):
             raise
+
